"""
A command to export Superset resources into a directory.
"""

from collections import defaultdict
from pathlib import Path
from zipfile import ZipFile

import click
import yaml
from yarl import URL

from preset_cli.api.clients.superset import SupersetClient
from preset_cli.lib import remove_root

JINJA2_OPEN_MARKER = "__JINJA2_OPEN__"
JINJA2_CLOSE_MARKER = "__JINJA2_CLOSE__"
assert JINJA2_OPEN_MARKER != JINJA2_CLOSE_MARKER


@click.command()
@click.argument("directory", type=click.Path(exists=True, resolve_path=True))
@click.option(
    "--overwrite",
    is_flag=True,
    default=False,
    help="Overwrite existing resources",
)
@click.pass_context
def export_assets(  # pylint: disable=too-many-locals
    ctx: click.core.Context,
    directory: str,
    overwrite: bool = False,
) -> None:
    """
    Export DBs/datasets/charts/dashboards to a directory.
    """
    auth = ctx.obj["AUTH"]
    url = URL(ctx.obj["INSTANCE"])
    client = SupersetClient(url, auth)
    root = Path(directory)

    for resource_name in ["database", "dataset", "chart", "dashboard"]:
        export_resource(resource_name, root, client, overwrite)


def export_resource(
    resource_name: str,
    root: Path,
    client: SupersetClient,
    overwrite: bool,
) -> None:
    """
    Export a given resource and unzip it in a directory.
    """
    resources = client.get_resources(resource_name)
    ids = [resource["id"] for resource in resources]
    buf = client.export_zip(resource_name, ids)

    with ZipFile(buf) as bundle:
        contents = {
            remove_root(file_name): bundle.read(file_name).decode()
            for file_name in bundle.namelist()
        }

    for file_name, file_contents in contents.items():
        # skip related files
        if not file_name.startswith(resource_name):
            continue

        target = root / file_name
        if target.exists() and not overwrite:
            raise Exception(
                f"File already exists and --overwrite was not specified: {target}",
            )
        if not target.parent.exists():
            target.parent.mkdir(parents=True, exist_ok=True)

        # escape any pre-existing Jinja2 templates
        file_contents = file_contents.replace(
            "{{",
            f"{JINJA2_OPEN_MARKER} '{{{{' {JINJA2_CLOSE_MARKER}",
        )
        file_contents = file_contents.replace(
            "}}",
            f"{JINJA2_OPEN_MARKER} '}}}}' {JINJA2_CLOSE_MARKER}",
        )
        file_contents = file_contents.replace(JINJA2_OPEN_MARKER, "{{")
        file_contents = file_contents.replace(JINJA2_CLOSE_MARKER, "}}")

        with open(target, "w", encoding="utf-8") as output:
            output.write(file_contents)


@click.command()
@click.argument(
    "path",
    type=click.Path(resolve_path=True),
    default="users.yaml",
)
@click.pass_context
def export_users(ctx: click.core.Context, path: str) -> None:
    """
    Export users and their roles to a YAML file.
    """
    auth = ctx.obj["AUTH"]
    url = URL(ctx.obj["INSTANCE"])
    client = SupersetClient(url, auth)

    users = [
        {k: v for k, v in user.items() if k != "id"} for user in client.export_users()
    ]

    with open(path, "w", encoding="utf-8") as output:
        yaml.dump(users, output)


@click.command()
@click.argument(
    "path",
    type=click.Path(resolve_path=True),
    default="rls.yaml",
)
@click.pass_context
def export_rls(ctx: click.core.Context, path: str) -> None:
    """
    Export RLS rules to a YAML file.
    """
    auth = ctx.obj["AUTH"]
    url = URL(ctx.obj["INSTANCE"])
    client = SupersetClient(url, auth)

    with open(path, "w", encoding="utf-8") as output:
        yaml.dump(list(client.export_rls()), output)


@click.command()
@click.argument(
    "path",
    type=click.Path(resolve_path=True),
    default="ownership.yaml",
)
@click.pass_context
def export_ownership(ctx: click.core.Context, path: str) -> None:
    """
    Export DBs/datasets/charts/dashboards ownership to a YAML file.
    """
    auth = ctx.obj["AUTH"]
    url = URL(ctx.obj["INSTANCE"])
    client = SupersetClient(url, auth)

    ownership = defaultdict(list)
    for resource_name in ["dataset", "chart", "dashboard"]:
        for resource in client.export_ownership(resource_name):
            ownership[resource_name].append(
                {
                    "name": resource["name"],
                    "uuid": str(resource["uuid"]),
                    "owners": resource["owners"],
                },
            )

    with open(path, "w", encoding="utf-8") as output:
        yaml.dump(dict(ownership), output)



"""
A command to sync dbt models/metrics to Superset and charts/dashboards back as exposures.
"""

import os.path
from pathlib import Path
from typing import Optional, Tuple

import click
import yaml
from marshmallow import EXCLUDE
from yarl import URL

from preset_cli.api.clients.dbt import DBTClient, MetricSchema, ModelSchema
from preset_cli.api.clients.superset import SupersetClient
from preset_cli.auth.token import TokenAuth
from preset_cli.cli.superset.sync.dbt.databases import sync_database
from preset_cli.cli.superset.sync.dbt.datasets import sync_datasets
from preset_cli.cli.superset.sync.dbt.exposures import sync_exposures
from preset_cli.cli.superset.sync.dbt.lib import apply_select
from preset_cli.exceptions import DatabaseNotFoundError


@click.command()
@click.argument("manifest", type=click.Path(exists=True, resolve_path=True))
@click.option("--project", help="Name of the dbt project", default="default")
@click.option("--target", help="Target name", default="dev")
@click.option(
    "--profiles",
    help="Location of profiles.yml file",
    type=click.Path(exists=True, resolve_path=True),
)
@click.option(
    "--exposures",
    help="Path to file where exposures will be written",
    type=click.Path(exists=False),
)
@click.option(
    "--import-db",
    is_flag=True,
    default=False,
    help="Import database to Superset",
)
@click.option(
    "--disallow-edits",
    is_flag=True,
    default=False,
    help="Mark resources as managed externally to prevent edits",
)
@click.option("--external-url-prefix", default="", help="Base URL for resources")
@click.option(
    "--select",
    "-s",
    help="Model selection",
    multiple=True,
)
@click.option(
    "--exclude",
    "-x",
    help="Models to exclude",
    multiple=True,
)
@click.pass_context
def dbt_core(  # pylint: disable=too-many-arguments, too-many-locals
    ctx: click.core.Context,
    manifest: str,
    project: str,
    target: str,
    select: Tuple[str, ...],
    exclude: Tuple[str, ...],
    profiles: Optional[str] = None,
    exposures: Optional[str] = None,
    import_db: bool = False,
    disallow_edits: bool = True,
    external_url_prefix: str = "",
) -> None:
    """
    Sync models/metrics from dbt Core to Superset and charts/dashboards to dbt exposures.
    """
    auth = ctx.obj["AUTH"]
    url = URL(ctx.obj["INSTANCE"])
    client = SupersetClient(url, auth)

    if profiles is None:
        profiles = os.path.expanduser("~/.dbt/profiles.yml")

    try:
        database = sync_database(
            client,
            Path(profiles),
            project,
            target,
            import_db,
            disallow_edits,
            external_url_prefix,
        )
    except DatabaseNotFoundError:
        click.echo("No database was found, pass --import-db to create")
        return

    with open(manifest, encoding="utf-8") as input_:
        configs = yaml.load(input_, Loader=yaml.SafeLoader)

    model_schema = ModelSchema()
    models = []
    for config in configs["nodes"].values():
        if config["resource_type"] == "model":
            # conform to the same schema that dbt Cloud uses for models
            config["uniqueId"] = config["unique_id"]
            models.append(model_schema.load(config, unknown=EXCLUDE))
    models = apply_select(models, select, exclude)

    metrics = []
    metric_schema = MetricSchema()
    for config in configs["metrics"].values():
        # conform to the same schema that dbt Cloud uses for metrics
        config["dependsOn"] = config["depends_on"]["nodes"]
        config["uniqueID"] = config["unique_id"]
        metrics.append(metric_schema.load(config, unknown=EXCLUDE))

    datasets = sync_datasets(
        client,
        models,
        metrics,
        database,
        disallow_edits,
        external_url_prefix,
    )
    if exposures:
        exposures = os.path.expanduser(exposures)
        sync_exposures(client, Path(exposures), datasets)


@click.command()
@click.argument("token")
@click.argument("job_id", type=click.INT)
@click.option(
    "--disallow-edits",
    is_flag=True,
    default=False,
    help="Mark resources as managed externally to prevent edits",
)
@click.option("--external-url-prefix", default="", help="Base URL for resources")
@click.option(
    "--select",
    "-s",
    help="Node selection (same syntax as dbt)",
    multiple=True,
)
@click.option(
    "--exclude",
    "-x",
    help="Models to exclude",
    multiple=True,
)
@click.pass_context
def dbt_cloud(  # pylint: disable=too-many-arguments, too-many-locals
    ctx: click.core.Context,
    token: str,
    job_id: int,
    select: Tuple[str, ...],
    exclude: Tuple[str, ...],
    disallow_edits: bool = True,
    external_url_prefix: str = "",
) -> None:
    """
    Sync models/metrics from dbt Cloud to Superset.
    """
    superset_auth = ctx.obj["AUTH"]
    url = URL(ctx.obj["INSTANCE"])
    superset_client = SupersetClient(url, superset_auth)

    dbt_auth = TokenAuth(token)
    dbt_client = DBTClient(dbt_auth)

    # with dbt cloud the database must already exist
    database_name = dbt_client.get_database_name(job_id)
    databases = superset_client.get_databases(database_name=database_name)
    if not databases:
        click.echo(f'No database named "{database_name}" was found')
        return
    if len(databases) > 1:
        raise Exception("More than one database with the same name found")

    database = databases[0]
    models = dbt_client.get_models(job_id)
    models = apply_select(models, select, exclude)
    metrics = dbt_client.get_metrics(job_id)

    sync_datasets(
        superset_client,
        models,
        metrics,
        database,
        disallow_edits,
        external_url_prefix,
    )



"""
Test authentication mechanisms.
"""

from pytest_mock import MockerFixture
from requests_mock.mocker import Mocker
from yarl import URL

from preset_cli.auth.main import Auth, UsernamePasswordAuth


def test_auth(mocker: MockerFixture) -> None:
    """
    Tests for the base class ``Auth``.
    """
    requests = mocker.patch("preset_cli.auth.main.requests")

    auth = Auth()
    assert auth.get_session() == requests.Session()


def test_username_password_auth(requests_mock: Mocker) -> None:
    """
    Tests for the username/password authentication mechanism.
    """
    csrf_token = "CSFR_TOKEN"
    requests_mock.get(
        "https://superset.example.org/login/",
        text=f'<html><body><input id="csrf_token" value="{csrf_token}"></body></html>',
    )
    requests_mock.post("https://superset.example.org/login/")

    auth = UsernamePasswordAuth(
        URL("https://superset.example.org/"),
        "admin",
        "password123",
    )
    assert auth.get_headers() == {
        "X-CSRFToken": csrf_token,
    }

    assert (
        requests_mock.last_request.text
        == "username=admin&password=password123&csrf_token=CSFR_TOKEN"
    )



"""
A simple client for interacting with the Preset API.
"""

from typing import Any, List, Union

from yarl import URL

from preset_cli import __version__
from preset_cli.auth.main import Auth
from preset_cli.lib import validate_response


class PresetClient:  # pylint: disable=too-few-public-methods

    """
    A client for the Preset API.
    """

    def __init__(self, baseurl: Union[str, URL], auth: Auth):
        # convert to URL if necessary
        self.baseurl = URL(baseurl)
        self.auth = auth
        self.auth.headers.update(
            {
                "User-Agent": "Preset CLI",
                "X-Client-Version": __version__,
            },
        )

    def get_teams(self) -> List[Any]:
        """
        Retrieve all teams based on membership.
        """
        session = self.auth.get_session()
        headers = self.auth.get_headers()
        response = session.get(self.baseurl / "api/v1/teams/", headers=headers)
        validate_response(response)

        payload = response.json()
        teams = payload["payload"]

        return teams

    def get_workspaces(self, team_name: str) -> List[Any]:
        """
        Retrieve all workspaces for a given team.
        """
        session = self.auth.get_session()
        headers = self.auth.get_headers()
        response = session.get(
            self.baseurl / "api/v1/teams" / team_name / "workspaces/",
            headers=headers,
        )
        validate_response(response)

        payload = response.json()
        workspaces = payload["payload"]

        return workspaces



"""
JWT auth.
"""

import yaml

from preset_cli.auth.lib import get_access_token, get_credentials_path
from preset_cli.auth.token import TokenAuth


class JWTAuth(TokenAuth):  # pylint: disable=too-few-public-methods
    """
    Auth via JWT.
    """

    @classmethod
    def from_stored_credentials(cls) -> "JWTAuth":
        """
        Build auth from stored credentials.
        """
        credentials_path = get_credentials_path()
        if not credentials_path.exists():
            raise Exception(f"Could not load credentials from {credentials_path}")

        with open(credentials_path, encoding="utf-8") as input_:
            credentials = yaml.load(input_, Loader=yaml.SafeLoader)

        token = get_access_token(**credentials)

        return JWTAuth(token)



"""
Run SQL queries on Superset.
"""
import os.path
import traceback
from operator import itemgetter
from pathlib import Path
from typing import List, Optional, Tuple

import click
from prompt_toolkit import PromptSession
from prompt_toolkit.completion import WordCompleter
from prompt_toolkit.history import FileHistory
from prompt_toolkit.lexers import PygmentsLexer
from prompt_toolkit.styles.pygments import style_from_pygments_cls
from pygments.lexers.sql import SqlLexer
from pygments.styles import get_style_by_name
from sqlparse.keywords import KEYWORDS
from tabulate import tabulate
from yarl import URL

from preset_cli.api.clients.superset import SupersetClient
from preset_cli.exceptions import SupersetError

sql_completer = WordCompleter(list(KEYWORDS))
style = style_from_pygments_cls(get_style_by_name("stata-dark"))


@click.command()
@click.option(
    "--database-id",
    default=None,
    help="Database ID (leave empty for options)",
    type=click.INT,
)
@click.option(
    "--schema",
    default=None,
    help="Schema",
)
@click.option("-e", "--execute", default=None, help="Run query non-interactively")
@click.pass_context
def sql(  # pylint: disable=too-many-arguments
    ctx: click.core.Context,
    database_id: Optional[int],
    schema: Optional[str] = None,
    execute: Optional[str] = None,
) -> None:
    """
    Run SQL against an Apache Superset database.
    """
    auth = ctx.obj["AUTH"]
    url = URL(ctx.obj["INSTANCE"])
    client = SupersetClient(url, auth)

    databases = client.get_databases()
    if not databases:
        click.echo("No databases available")
        return None

    if len(databases) == 1 and database_id is None:
        database_id = databases[0]["id"]

    if database_id is None:
        click.echo("Choose the ID of a database to connect to:")
        for database in sorted(databases, key=itemgetter("id")):
            click.echo(f'({database["id"]}) {database["database_name"]}')
    while database_id is None:
        try:
            choice = int(input("> "))
            if any(database["id"] == choice for database in databases):
                database_id = choice
                break
        except ValueError:
            pass
        click.echo("Invalid choice")

    database_name = [
        database["database_name"]
        for database in databases
        if database["id"] == database_id
    ][0]

    if execute:
        return run_query(client, database_id, schema, execute)

    return run_session(client, database_id, database_name, schema, url)


def run_query(
    client: SupersetClient,
    database_id: int,
    schema: Optional[str],
    query: str,
) -> None:
    """
    Run a query in a given database.
    """
    try:
        results = client.run_query(database_id, query, schema)
        click.echo(tabulate(results, headers=results.columns, showindex=False))
    except SupersetError as ex:
        click.echo(
            click.style(
                "\n".join(error["message"] for error in ex.errors),
                fg="bright_red",
            ),
        )
    except Exception:  # pylint: disable=broad-except
        traceback.print_exc()


def run_session(
    client: SupersetClient,
    database_id: int,
    database_name: str,
    schema: Optional[str],
    url: URL,
) -> None:
    """
    Run SQL queries in an interactive session.
    """
    history = Path(os.path.expanduser("~/.config/preset-cli/"))
    if not history.exists():
        history.mkdir(parents=True)

    session = PromptSession(
        lexer=PygmentsLexer(SqlLexer),
        completer=sql_completer,
        style=style,
        history=FileHistory(history / f"sql-{url.host}-{database_id}.history"),
    )

    lines: List[str] = []
    quote_context = " "
    padding = " " * (len(database_name) - 1)
    while True:
        prompt = f"{database_name}> " if not lines else f"{padding}{quote_context}. "
        try:
            line = session.prompt(prompt)
        except KeyboardInterrupt:
            lines = []
            quote_context = " "
            continue  # Control-C pressed. Try again.
        except EOFError:
            break  # Control-D pressed.

        lines.append(line)
        query = "\n".join(lines)

        is_terminated, quote_context = get_query_termination(query)
        if is_terminated:
            run_query(client, database_id, schema, query)
            lines = []
            quote_context = " "

    click.echo("Goodbye!")


def get_query_termination(query: str) -> Tuple[bool, str]:
    """
    Check if a query is ended or if a new line should be created.

    This function looks for a semicolon at the end, making sure no quotation mark must be
    closed.
    """
    quote_context = " "
    quote_chars = ('"', "'", "`")

    for query_char in query:
        if quote_context == query_char:
            quote_context = " "
        else:
            for quote in quote_chars:
                if quote_context == " " and quote == query_char:
                    quote_context = quote

    is_terminated = quote_context == " " and query.endswith(";")

    return is_terminated, quote_context



"""
Tests for the Superset dispatcher.
"""

import click
from click.testing import CliRunner

from preset_cli.cli.superset.main import mutate_commands, superset


def test_mutate_commands() -> None:
    """
    Test ``mutate_commands``.
    """

    @click.group()
    def source_group() -> None:
        """
        A simple group of commands.
        """

    @click.command()
    @click.argument("name")
    def source_command(name: str) -> None:
        """
        Say hello.
        """
        click.echo(f"Hello, {name}!")

    source_group.add_command(source_command)

    @click.group()
    def source_subgroup() -> None:
        """
        A simple subgroup.
        """

    @click.command()
    @click.argument("name")
    def source_subcommand(name: str) -> None:
        """
        Say goodbye
        """
        click.echo(f"Goodbye, {name}!")

    source_subgroup.add_command(source_subcommand)
    source_group.add_command(source_subgroup)

    @click.group()
    @click.pass_context
    def target_group(ctx: click.core.Context) -> None:
        """
        The target group to which commands will be added to.
        """
        ctx.ensure_object(dict)
        ctx.obj["WORKSPACES"] = ["instance1", "instance2"]

    mutate_commands(source_group, target_group)

    runner = CliRunner()

    result = runner.invoke(
        target_group,
        ["source-command", "Alice"],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    assert (
        result.output
        == """
instance1
Hello, Alice!

instance2
Hello, Alice!
"""
    )

    result = runner.invoke(
        target_group,
        ["source-subgroup", "source-subcommand", "Alice"],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    assert (
        result.output
        == """
instance1
Goodbye, Alice!

instance2
Goodbye, Alice!
"""
    )


def test_superset() -> None:
    """
    Test the ``superset`` command.
    """
    runner = CliRunner()

    result = runner.invoke(superset, ["--help"], catch_exceptions=False)
    assert result.exit_code == 0
    assert (
        result.output
        == """Usage: superset [OPTIONS] COMMAND [ARGS]...

  Send commands to one or more Superset instances.

Options:
  --help  Show this message and exit.

Commands:
  export
  export-assets
  export-ownership
  export-rls
  export-users
  import-assets
  import-ownership
  import-rls
  sql
  sync
"""
    )

    result = runner.invoke(superset, ["export", "--help"], catch_exceptions=False)
    assert result.exit_code == 0
    assert (
        result.output
        == """Usage: superset export [OPTIONS] DIRECTORY

Options:
  --overwrite  Overwrite existing resources
  --help       Show this message and exit.
"""
    )



"""
Tests for ``preset_cli.cli.main``.
"""
# pylint: disable=unused-argument, invalid-name, redefined-outer-name

from pathlib import Path
from typing import Any, Dict

import pytest
import yaml
from click.testing import CliRunner
from pyfakefs.fake_filesystem import FakeFilesystem
from pytest_mock import MockerFixture
from yarl import URL

from preset_cli.cli.main import (
    get_status_icon,
    parse_workspace_selection,
    preset_cli,
    split_comma,
)


def test_split_comma(mocker: MockerFixture) -> None:
    """
    Test ``split_comma``.

    This is used to split workspaces passed in the CLI.
    """
    ctx = mocker.MagicMock()
    assert split_comma(
        ctx,
        "workspaces",
        "https://ws1.preset.io/,https://ws3.preset.io/, https://ws2.preset.io/",
    ) == [
        "https://ws1.preset.io/",
        "https://ws3.preset.io/",
        "https://ws2.preset.io/",
    ]
    assert split_comma(ctx, "workspaces", None) == []


def test_get_status_icon() -> None:
    """
    Test ``get_status_icon``.
    """
    assert get_status_icon("READY") == "✅"
    assert get_status_icon("LOADING_EXAMPLES") == "📊"
    assert get_status_icon("CREATING_DB") == "💾"
    assert get_status_icon("INITIALIZING_DB") == "💾"
    assert get_status_icon("MIGRATING_DB") == "🚧"
    assert get_status_icon("ROTATING_SECRETS") == "🕵️"
    assert get_status_icon("UNKNOWN") == "❓"
    assert get_status_icon("ERROR") == "❗️"
    assert get_status_icon("UPGRADING") == "⤴️"
    assert get_status_icon("INVALID") == "❓"


def test_parse_workspace_selection() -> None:
    """
    Test ``parse_workspace_selection``.
    """
    assert parse_workspace_selection("1-4,7", 10) == [1, 2, 3, 4, 7]
    assert parse_workspace_selection("-4", 10) == [1, 2, 3, 4]
    assert parse_workspace_selection("4-", 10) == [4, 5, 6, 7, 8, 9, 10]

    with pytest.raises(Exception) as excinfo:
        parse_workspace_selection("1-20", 10)
    assert str(excinfo.value) == "End 20 is greater than 10"
    with pytest.raises(Exception) as excinfo:
        parse_workspace_selection("20", 10)
    assert str(excinfo.value) == "Number 20 is greater than 10"


def test_auth(mocker: MockerFixture) -> None:
    """
    Test the ``auth`` command.
    """
    credentials_path = Path("/path/to/config/credentials.yaml")
    mocker.patch(
        "preset_cli.cli.main.get_credentials_path",
        return_value=credentials_path,
    )
    webbrowser = mocker.patch("preset_cli.cli.main.webbrowser")
    mocker.patch("preset_cli.cli.main.input", return_value="API_TOKEN")
    getpass = mocker.patch("preset_cli.cli.main.getpass")
    getpass.getpass.return_value = "API_SECRET"
    store_credentials = mocker.patch("preset_cli.cli.main.store_credentials")

    runner = CliRunner()
    result = runner.invoke(
        preset_cli,
        ["--jwt-token", "JWT_TOKEN", "auth"],
        catch_exceptions=False,
    )
    assert result.exit_code == 0

    webbrowser.open.assert_called_with("https://manage.app.preset.io/app/user")
    store_credentials.assert_called_with(
        "API_TOKEN",
        "API_SECRET",
        URL("https://manage.app.preset.io/"),
        credentials_path,
    )


def test_auth_show(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test the ``auth --show`` command.
    """
    credentials_path = Path("/path/to/config/credentials.yaml")
    fs.create_file(
        credentials_path,
        contents=yaml.dump(
            {
                "baseurl": "https://manage.app.preset.io/",
                "api_secret": "XXX",
                "api_token": "abc",
            },
        ),
    )
    mocker.patch(
        "preset_cli.cli.main.get_credentials_path",
        return_value=credentials_path,
    )

    runner = CliRunner()
    result = runner.invoke(
        preset_cli,
        ["--jwt-token", "JWT_TOKEN", "auth", "--show"],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    assert (
        result.output
        == """/path/to/config/credentials.yaml
================================
api_secret: XXX
api_token: abc
baseurl: https://manage.app.preset.io/

"""
    )


def test_auth_show_no_credentials(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test the ``auth --show`` command when there are no credentials.
    """
    credentials_path = Path("/path/to/config/credentials.yaml")
    mocker.patch(
        "preset_cli.cli.main.get_credentials_path",
        return_value=credentials_path,
    )

    runner = CliRunner()
    result = runner.invoke(
        preset_cli,
        ["--jwt-token", "JWT_TOKEN", "auth", "--show"],
        catch_exceptions=False,
    )
    assert result.exit_code == 1
    assert result.output == (
        "The file /path/to/config/credentials.yaml doesn't exist. "
        "Run ``preset-cli auth`` to create it.\n"
    )


def test_auth_overwrite(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test the ``auth`` command when credentials already exist.
    """
    credentials_path = Path("/path/to/config/credentials.yaml")
    fs.create_file(credentials_path)
    mocker.patch(
        "preset_cli.cli.main.get_credentials_path",
        return_value=credentials_path,
    )

    runner = CliRunner()

    result = runner.invoke(
        preset_cli,
        ["--jwt-token", "JWT_TOKEN", "auth"],
        catch_exceptions=False,
    )
    assert result.exit_code == 1

    mocker.patch("preset_cli.cli.main.webbrowser")
    mocker.patch("preset_cli.cli.main.input", return_value="API_TOKEN")
    getpass = mocker.patch("preset_cli.cli.main.getpass")
    getpass.getpass.return_value = "API_SECRET"
    mocker.patch("preset_cli.cli.main.store_credentials")
    result = runner.invoke(
        preset_cli,
        ["--jwt-token", "JWT_TOKEN", "auth", "--overwrite"],
        catch_exceptions=False,
    )
    assert result.exit_code == 0


def test_auth_overwrite_expired_credentials(
    mocker: MockerFixture,
    fs: FakeFilesystem,
) -> None:
    """
    Test the ``auth`` command when overwriting expired credentials.
    """
    credentials_path = Path("/path/to/config/credentials.yaml")
    fs.create_file(
        credentials_path,
        contents=yaml.dump({"api_secret": "API_SECRET", "api_token": "API_TOKEN"}),
    )
    mocker.patch(
        "preset_cli.cli.main.get_credentials_path",
        return_value=credentials_path,
    )
    get_access_token = mocker.patch(
        "preset_cli.cli.main.get_access_token",
        side_effect=Exception("Unable to get access token"),
    )

    runner = CliRunner()

    mocker.patch("preset_cli.cli.main.webbrowser")
    mocker.patch("preset_cli.cli.main.input", return_value="API_TOKEN")
    getpass = mocker.patch("preset_cli.cli.main.getpass")
    getpass.getpass.return_value = "API_SECRET"
    mocker.patch("preset_cli.cli.main.store_credentials")
    result = runner.invoke(
        preset_cli,
        ["auth", "--overwrite"],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    get_access_token.assert_called_with(
        URL("https://manage.app.preset.io/"),
        "API_TOKEN",
        "API_SECRET",
    )


def test_jwt_token_credentials_exist(
    mocker: MockerFixture,
    fs: FakeFilesystem,
) -> None:
    """
    Test the command when the credentials are stored.
    """
    credentials_path = Path("/path/to/config/credentials.yaml")
    fs.create_file(
        credentials_path,
        contents=yaml.dump({"api_secret": "API_SECRET", "api_token": "API_TOKEN"}),
    )
    mocker.patch(
        "preset_cli.cli.main.get_credentials_path",
        return_value=credentials_path,
    )
    mocker.patch("preset_cli.cli.main.get_access_token", return_value="JWT_TOKEN")
    JWTAuth = mocker.patch("preset_cli.cli.main.JWTAuth")

    runner = CliRunner()
    result = runner.invoke(preset_cli, ["auth", "--help"], catch_exceptions=False)
    assert result.exit_code == 0
    JWTAuth.assert_called_with("JWT_TOKEN")


def test_jwt_token_invalid_credentials(
    mocker: MockerFixture,
    fs: FakeFilesystem,
) -> None:
    """
    Test the command when the credentials are stored.
    """
    credentials_path = Path("/path/to/config/credentials.yaml")
    fs.create_file(
        credentials_path,
        contents=yaml.dump({"api_token": "API_TOKEN"}),
    )
    mocker.patch(
        "preset_cli.cli.main.get_credentials_path",
        return_value=credentials_path,
    )
    mocker.patch("preset_cli.cli.main.get_access_token", return_value="JWT_TOKEN")

    runner = CliRunner()
    result = runner.invoke(preset_cli, ["auth", "--help"], catch_exceptions=False)
    assert result.exit_code == 1


def test_jwt_token_prompt_for_credentials(
    mocker: MockerFixture,
    fs: FakeFilesystem,
) -> None:
    """
    Test the command when the credentials are stored.
    """
    credentials_path = Path("/path/to/config/credentials.yaml")
    mocker.patch(
        "preset_cli.cli.main.get_credentials_path",
        return_value=credentials_path,
    )
    mocker.patch("preset_cli.cli.main.webbrowser")
    mocker.patch("preset_cli.cli.main.input", return_value="API_TOKEN")
    getpass = mocker.patch("preset_cli.cli.main.getpass")
    getpass.getpass.return_value = "API_SECRET"
    mocker.patch("preset_cli.cli.main.store_credentials")
    mocker.patch("preset_cli.cli.main.get_access_token", return_value="JWT_TOKEN")
    JWTAuth = mocker.patch("preset_cli.cli.main.JWTAuth")

    runner = CliRunner()
    result = runner.invoke(preset_cli, ["auth", "--help"], catch_exceptions=False)
    assert result.exit_code == 0
    JWTAuth.assert_called_with("JWT_TOKEN")


def test_jwt_token_credentials_passed(
    mocker: MockerFixture,
    fs: FakeFilesystem,
) -> None:
    """
    Test the command when the credentials are stored.
    """
    mocker.patch("preset_cli.cli.main.get_access_token", return_value="JWT_TOKEN")
    JWTAuth = mocker.patch("preset_cli.cli.main.JWTAuth")

    runner = CliRunner()
    result = runner.invoke(
        preset_cli,
        ["--api-token", "API_TOKEN", "--api-secret", "API_SECRET", "auth", "--help"],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    JWTAuth.assert_called_with("JWT_TOKEN")


def test_workspaces(mocker: MockerFixture) -> None:
    """
    Test that we prompt user for their workspaces if not specified.
    """
    PresetClient = mocker.patch("preset_cli.cli.main.PresetClient")
    client = PresetClient()
    client.get_teams.return_value = [{"name": "botafogo", "title": "Alvinegro"}]
    client.get_workspaces.return_value = [
        {"workspace_status": "READY", "title": "My Workspace", "hostname": "ws1"},
        {"workspace_status": "READY", "title": "My Other Workspace", "hostname": "ws2"},
    ]
    mocker.patch("preset_cli.cli.main.input", side_effect=["invalid", "-"])

    runner = CliRunner()
    obj: Dict[str, Any] = {}
    result = runner.invoke(
        preset_cli,
        ["--jwt-token", "JWT_TOKEN", "superset", "--help"],
        catch_exceptions=False,
        obj=obj,
    )
    assert result.exit_code == 0
    assert obj["WORKSPACES"] == ["https://ws1", "https://ws2"]


def test_workspaces_since_workspace(mocker: MockerFixture) -> None:
    """
    Test that we don't prompt user for their workspaces if they have only one.
    """
    PresetClient = mocker.patch("preset_cli.cli.main.PresetClient")
    client = PresetClient()
    client.get_teams.return_value = [{"name": "botafogo", "title": "Alvinegro"}]
    client.get_workspaces.return_value = [
        {"workspace_status": "READY", "title": "My Workspace", "hostname": "ws1"},
    ]
    parse_workspace_selection = mocker.patch(
        "preset_cli.cli.main.parse_workspace_selection",
    )

    runner = CliRunner()
    obj: Dict[str, Any] = {}
    result = runner.invoke(
        preset_cli,
        ["--jwt-token", "JWT_TOKEN", "superset", "--help"],
        catch_exceptions=False,
        obj=obj,
    )
    assert result.exit_code == 0
    assert obj["WORKSPACES"] == ["https://ws1"]
    parse_workspace_selection.assert_not_called()


def test_workspaces_no_workspaces(mocker: MockerFixture) -> None:
    """
    Test when no workspaces are available.
    """
    PresetClient = mocker.patch("preset_cli.cli.main.PresetClient")
    client = PresetClient()
    client.get_teams.return_value = [{"name": "botafogo", "title": "Alvinegro"}]
    client.get_workspaces.return_value = []
    mocker.patch("preset_cli.cli.main.input", side_effect=["invalid", "-"])

    runner = CliRunner()
    obj: Dict[str, Any] = {}
    result = runner.invoke(
        preset_cli,
        ["--jwt-token", "JWT_TOKEN", "superset", "--help"],
        catch_exceptions=False,
        obj=obj,
    )
    assert result.exit_code == 1



"""
A command to sync Superset exports into a Superset instance.
"""

import getpass
import importlib.util
import os
from datetime import datetime, timezone
from io import BytesIO
from pathlib import Path
from types import ModuleType
from typing import Any, Dict, Tuple
from zipfile import ZipFile

import click
import yaml
from jinja2 import Template
from sqlalchemy.engine.url import make_url
from yarl import URL

from preset_cli.api.clients.superset import SupersetClient
from preset_cli.exceptions import SupersetError

YAML_EXTENSIONS = {".yaml", ".yml"}

# This should be identical to ``superset.models.core.PASSWORD_MASK``. It's duplicated here
# because we don't want to have the CLI to depend on the ``superset`` package.
PASSWORD_MASK = "X" * 10


resource_types = {
    "chart": "Slice",
    "dashboard": "Dashboard",
    "database": "Database",
    "dataset": "SqlaTable",
}


def load_user_modules(root: Path) -> Dict[str, ModuleType]:
    """
    Load user-defined modules so they can be used with Jinja2.
    """
    modules = {}
    for path in root.glob("*.py"):
        spec = importlib.util.spec_from_file_location(path.stem, path)
        if spec and spec.loader:
            modules[path.stem] = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(modules[path.stem])  # type: ignore

    return modules


def raise_helper(message: str, *args: Any) -> None:
    """
    Macro for Jinja2 so users can raise exceptions.
    """
    raise Exception(message % args)


@click.command()
@click.argument("directory", type=click.Path(exists=True, resolve_path=True))
@click.option(
    "--overwrite",
    is_flag=True,
    default=False,
    help="Overwrite existing resources",
)
@click.option(
    "--option",
    "-o",
    multiple=True,
    help="Custom values for templates (eg, country=BR)",
)
@click.option(
    "--disallow-edits",
    is_flag=True,
    default=False,
    help="Mark resources as manged externally to prevent edits",
)
@click.option("--external-url-prefix", default="", help="Base URL for resources")
@click.option(
    "--load-env",
    "-e",
    is_flag=True,
    default=False,
    help="Load environment variables to ``env[]`` template helper",
)
@click.pass_context
def native(  # pylint: disable=too-many-locals, too-many-arguments
    ctx: click.core.Context,
    directory: str,
    option: Tuple[str, ...],
    overwrite: bool = False,
    disallow_edits: bool = True,  # pylint: disable=unused-argument
    external_url_prefix: str = "",
    load_env: bool = False,
) -> None:
    """
    Sync exported DBs/datasets/charts/dashboards to Superset.
    """
    auth = ctx.obj["AUTH"]
    url = URL(ctx.obj["INSTANCE"])
    client = SupersetClient(url, auth)
    root = Path(directory)

    base_url = URL(external_url_prefix) if external_url_prefix else None

    # env for Jinja2 templating
    env = dict(pair.split("=", 1) for pair in option if "=" in pair)  # type: ignore
    env["instance"] = url
    env["functions"] = load_user_modules(root / "functions")
    env["raise"] = raise_helper
    if load_env:
        env["env"] = os.environ

    # read all the YAML files
    contents: Dict[str, str] = {}
    queue = [root]
    while queue:
        path_name = queue.pop()
        if path_name.is_dir() and not path_name.stem.startswith("."):
            queue.extend(path_name.glob("*"))
        elif path_name.suffix.lower() in YAML_EXTENSIONS:
            with open(path_name, encoding="utf-8") as input_:
                template = Template(input_.read())
                content = template.render(**env)
                relative_path = path_name.relative_to(root)

                # mark resource as being managed externally
                config = yaml.load(content, Loader=yaml.SafeLoader)
                config["is_managed_externally"] = disallow_edits
                if base_url:
                    config["external_url"] = str(
                        base_url / str(relative_path),
                    )
                if relative_path.parts[0] == "databases":
                    prompt_for_passwords(relative_path, config)

                contents[str("bundle" / relative_path)] = yaml.safe_dump(config)

    # TODO (betodealmeida): use endpoint from https://github.com/apache/superset/pull/19220
    for resource in ["database", "dataset", "chart", "dashboard"]:
        import_resource(resource, contents, client, overwrite)


def prompt_for_passwords(path: Path, config: Dict[str, Any]) -> None:
    """
    Prompt user for masked passwords.

    Modify the config in place.
    """
    uri = config["sqlalchemy_uri"]
    password = make_url(uri).password
    if password == PASSWORD_MASK and config.get("password") is None:
        config["password"] = getpass.getpass(
            f"Please provide the password for {path}: ",
        )


def import_resource(
    resource: str,
    contents: Dict[str, str],
    client: SupersetClient,
    overwrite: bool,
) -> None:
    """
    Import a given resource.
    """
    contents["bundle/metadata.yaml"] = yaml.dump(
        dict(
            version="1.0.0",
            type=resource_types[resource],
            timestamp=datetime.now(tz=timezone.utc).isoformat(),
        ),
    )

    buf = BytesIO()
    with ZipFile(buf, "w") as bundle:
        for file_path, file_content in contents.items():
            with bundle.open(file_path, "w") as output:
                output.write(file_content.encode())
    buf.seek(0)
    try:
        client.import_zip(resource, buf, overwrite=overwrite)
    except SupersetError as ex:
        click.echo(
            click.style(
                "\n".join(error["message"] for error in ex.errors),
                fg="bright_red",
            ),
        )

        # check if overwrite is needed:
        existing = [
            key
            for error in ex.errors
            for key, value in error.get("extra", {}).items()
            if "overwrite=true" in value
        ]
        if not existing:
            raise ex

        existing_list = "\n".join("- " + name for name in existing)
        click.echo(
            click.style(
                (
                    "The following file(s) already exist. Pass --overwrite to "
                    f"replace them.\n{existing_list}"
                ),
                fg="bright_red",
            ),
        )



"""
Test the ``sql`` command.
"""
# pylint: disable=invalid-name, unused-argument, redefined-outer-name

from io import StringIO
from pathlib import Path

import pandas as pd
from click.testing import CliRunner
from pyfakefs.fake_filesystem import FakeFilesystem
from pytest_mock import MockerFixture
from yarl import URL

from preset_cli.cli.superset.main import superset_cli
from preset_cli.cli.superset.sql import run_query, run_session
from preset_cli.exceptions import ErrorLevel, SupersetError


def test_run_query(mocker: MockerFixture) -> None:
    """
    Test ``run_query``.
    """
    client = mocker.MagicMock()
    client.run_query.return_value = pd.DataFrame([{"answer": 42}])
    click = mocker.patch("preset_cli.cli.superset.sql.click")

    run_query(client=client, database_id=1, schema=None, query="SELECT 42 AS answer")
    client.run_query.assert_called_with(1, "SELECT 42 AS answer", None)
    click.echo.assert_called_with("  answer\n--------\n      42")


def test_run_query_superset_error(mocker: MockerFixture) -> None:
    """
    Test ``run_query`` when a ``SupersetError`` happens.
    """
    client = mocker.MagicMock()
    client.run_query.side_effect = SupersetError(
        [
            {
                "message": "Only SELECT statements are allowed against this database.",
                "error_type": "DML_NOT_ALLOWED_ERROR",
                "level": ErrorLevel.ERROR,
                "extra": {
                    "issue_codes": [
                        {
                            "code": 1022,
                            "message": "Issue 1022 - Database does not allow data manipulation.",
                        },
                    ],
                },
            },
        ],
    )
    click = mocker.patch("preset_cli.cli.superset.sql.click")

    run_query(client=client, database_id=1, schema=None, query="SSELECT 1")
    click.style.assert_called_with(
        "Only SELECT statements are allowed against this database.",
        fg="bright_red",
    )


def test_run_query_exception(mocker: MockerFixture) -> None:
    """
    Test ``run_query`` when a different exception happens.
    """
    client = mocker.MagicMock()
    client.run_query.side_effect = Exception("Unexpected error")
    traceback = mocker.patch("preset_cli.cli.superset.sql.traceback")

    run_query(client=client, database_id=1, schema=None, query="SSELECT 1")
    traceback.print_exc.assert_called_with()


def test_run_session(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test ``run_session``.
    """
    history = Path("/path/to/.config/preset-cli/")
    os = mocker.patch("preset_cli.cli.superset.sql.os")
    os.path.expanduser.return_value = str(history)

    client = mocker.MagicMock()
    client.run_query.return_value = pd.DataFrame([{"answer": 42}])

    stdout = mocker.patch("sys.stdout", new_callable=StringIO)
    PromptSession = mocker.patch("preset_cli.cli.superset.sql.PromptSession")
    session = PromptSession()
    session.prompt.side_effect = ["SELECT 42 AS answer;", "", EOFError()]

    run_session(
        client=client,
        database_id=1,
        database_name="GSheets",
        schema=None,
        url=URL("https://superset.example.org/"),
    )
    result = stdout.getvalue()
    assert (
        result
        == """  answer
--------
      42
Goodbye!
"""
    )


def test_run_session_multiple_commands(
    mocker: MockerFixture,
    fs: FakeFilesystem,
) -> None:
    """
    Test ``run_session``.
    """
    history = Path("/path/to/.config/preset-cli/")
    os = mocker.patch("preset_cli.cli.superset.sql.os")
    os.path.expanduser.return_value = str(history)

    client = mocker.MagicMock()
    client.run_query.side_effect = [
        pd.DataFrame([{"answer": 42}]),
        pd.DataFrame([{"question": "Life, universe, everything"}]),
    ]

    stdout = mocker.patch("sys.stdout", new_callable=StringIO)
    PromptSession = mocker.patch("preset_cli.cli.superset.sql.PromptSession")
    session = PromptSession()
    session.prompt.side_effect = [
        "SELECT 42 AS answer;",
        "SELECT 'Life, universe, everything' AS question;",
        "",
        EOFError(),
    ]

    run_session(
        client=client,
        database_id=1,
        database_name="GSheets",
        schema=None,
        url=URL("https://superset.example.org/"),
    )
    result = stdout.getvalue()
    assert (
        result
        == """  answer
--------
      42
question
--------------------------
Life, universe, everything
Goodbye!
"""
    )


def test_run_session_multiline(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test ``run_session`` with multilines.
    """
    history = Path("/path/to/.config/preset-cli/")
    os = mocker.patch("preset_cli.cli.superset.sql.os")
    os.path.expanduser.return_value = str(history)

    client = mocker.MagicMock()
    client.run_query.return_value = pd.DataFrame([{"the\nanswer": "foo\nbar"}])

    stdout = mocker.patch("sys.stdout", new_callable=StringIO)
    PromptSession = mocker.patch("preset_cli.cli.superset.sql.PromptSession")
    session = PromptSession()
    session.prompt.side_effect = [
        """SELECT 'foo\nbar' AS "the\nanswer";""",
        "",
        EOFError(),
    ]

    run_session(
        client=client,
        database_id=1,
        database_name="GSheets",
        schema=None,
        url=URL("https://superset.example.org/"),
    )
    result = stdout.getvalue()
    assert (
        result
        == """the
answer
--------
foo
bar
Goodbye!
"""
    )


def test_run_session_ctrl_c(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test that ``CTRL-C`` does not exit the REPL.
    """
    history = Path("/path/to/.config/preset-cli/")
    os = mocker.patch("preset_cli.cli.superset.sql.os")
    os.path.expanduser.return_value = str(history)

    client = mocker.MagicMock()
    client.run_query.return_value = pd.DataFrame([{"answer": 42}])

    stdout = mocker.patch("sys.stdout", new_callable=StringIO)
    PromptSession = mocker.patch("preset_cli.cli.superset.sql.PromptSession")
    session = PromptSession()
    session.prompt.side_effect = [KeyboardInterrupt(), "SELECT 1;", EOFError()]

    run_session(
        client=client,
        database_id=1,
        database_name="GSheets",
        schema=None,
        url=URL("https://superset.example.org/"),
    )
    result = stdout.getvalue()
    assert (
        result
        == """  answer
--------
      42
Goodbye!
"""
    )


def test_run_session_history_exists(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test ``run_session``.
    """
    history = Path("/path/to/.config/preset-cli/")
    os = mocker.patch("preset_cli.cli.superset.sql.os")
    os.path.expanduser.return_value = str(history)
    history.mkdir(parents=True)

    client = mocker.MagicMock()
    client.run_query.return_value = pd.DataFrame([{"answer": 42}])

    stdout = mocker.patch("sys.stdout", new_callable=StringIO)
    PromptSession = mocker.patch("preset_cli.cli.superset.sql.PromptSession")
    session = PromptSession()
    session.prompt.side_effect = ["SELECT 42 AS answer;", "", EOFError()]

    run_session(
        client=client,
        database_id=1,
        database_name="GSheets",
        schema=None,
        url=URL("https://superset.example.org/"),
    )
    result = stdout.getvalue()
    assert (
        result
        == """  answer
--------
      42
Goodbye!
"""
    )


def test_sql_run_query(mocker: MockerFixture) -> None:
    """
    Test the ``sql`` command in programmatic mode (run single query).
    """
    SupersetClient = mocker.patch("preset_cli.cli.superset.sql.SupersetClient")
    client = SupersetClient()
    client.get_databases.return_value = [{"id": 1, "database_name": "GSheets"}]
    mocker.patch("preset_cli.cli.superset.main.UsernamePasswordAuth")
    run_query = mocker.patch("preset_cli.cli.superset.sql.run_query")

    runner = CliRunner()
    result = runner.invoke(
        superset_cli,
        [
            "https://superset.example.org/",
            "sql",
            "-e",
            "SELECT 1",
            "--database-id",
            "1",
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    run_query.assert_called_with(client, 1, None, "SELECT 1")


def test_sql_run_session(mocker: MockerFixture) -> None:
    """
    Test the ``sql`` command in session mode (REPL).
    """
    SupersetClient = mocker.patch("preset_cli.cli.superset.sql.SupersetClient")
    client = SupersetClient()
    client.get_databases.return_value = [{"id": 1, "database_name": "GSheets"}]
    mocker.patch("preset_cli.cli.superset.main.UsernamePasswordAuth")
    run_session = mocker.patch("preset_cli.cli.superset.sql.run_session")

    runner = CliRunner()
    result = runner.invoke(
        superset_cli,
        [
            "https://superset.example.org/",
            "sql",
            "--database-id",
            "1",
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    run_session.assert_called_with(
        client,
        1,
        "GSheets",
        None,
        URL("https://superset.example.org/"),
    )


def test_sql_run_query_no_databases(mocker: MockerFixture) -> None:
    """
    Test the ``sql`` command when no databases are found.
    """
    SupersetClient = mocker.patch("preset_cli.cli.superset.sql.SupersetClient")
    client = SupersetClient()
    client.get_databases.return_value = []
    mocker.patch("preset_cli.cli.superset.main.UsernamePasswordAuth")
    mocker.patch("preset_cli.cli.superset.sql.run_query")

    runner = CliRunner()
    result = runner.invoke(
        superset_cli,
        [
            "https://superset.example.org/",
            "sql",
            "-e",
            "SELECT 1",
            "--database-id",
            "1",
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    assert result.output == "No databases available\n"


def test_sql_choose_database(mocker: MockerFixture) -> None:
    """
    Test the ``sql`` command choosing a DB interactively.
    """
    SupersetClient = mocker.patch("preset_cli.cli.superset.sql.SupersetClient")
    client = SupersetClient()
    client.get_databases.return_value = [
        {"id": 1, "database_name": "GSheets"},
        {"id": 2, "database_name": "Trino"},
    ]
    mocker.patch("preset_cli.cli.superset.main.UsernamePasswordAuth")
    mocker.patch("preset_cli.cli.superset.sql.input", side_effect=["3", "invalid", "1"])
    run_query = mocker.patch("preset_cli.cli.superset.sql.run_query")

    runner = CliRunner()
    result = runner.invoke(
        superset_cli,
        [
            "https://superset.example.org/",
            "sql",
            "-e",
            "SELECT 1",
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    run_query.assert_called_with(client, 1, None, "SELECT 1")


def test_sql_single_database(mocker: MockerFixture) -> None:
    """
    Test the ``sql`` command when there's a single database available.
    """
    SupersetClient = mocker.patch("preset_cli.cli.superset.sql.SupersetClient")
    client = SupersetClient()
    client.get_databases.return_value = [
        {"id": 1, "database_name": "GSheets"},
    ]
    mocker.patch("preset_cli.cli.superset.main.UsernamePasswordAuth")
    run_query = mocker.patch("preset_cli.cli.superset.sql.run_query")

    runner = CliRunner()
    result = runner.invoke(
        superset_cli,
        [
            "https://superset.example.org/",
            "sql",
            "-e",
            "SELECT 1",
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    run_query.assert_called_with(client, 1, None, "SELECT 1")



"""
Token auth.
"""

from typing import Dict

from preset_cli.auth.main import Auth


class TokenAuth(Auth):  # pylint: disable=too-few-public-methods
    """
    Auth via a token.
    """

    def __init__(self, token: str):
        super().__init__()
        self.token = token

    def get_headers(self) -> Dict[str, str]:
        return {"Authorization": f"Bearer {self.token}"}



"""
Sync dbt datasets/etrics to Superset.
"""

# pylint: disable=consider-using-f-string

import json
import logging
from typing import Any, List

from yarl import URL

from preset_cli.api.clients.dbt import MetricSchema, ModelSchema
from preset_cli.api.clients.superset import SupersetClient
from preset_cli.api.operators import OneToMany
from preset_cli.cli.superset.sync.dbt.metrics import get_metric_expression

_logger = logging.getLogger(__name__)


def sync_datasets(  # pylint: disable=too-many-locals, too-many-branches, too-many-arguments
    client: SupersetClient,
    models: List[ModelSchema],
    metrics: List[MetricSchema],
    database: Any,
    disallow_edits: bool,
    external_url_prefix: str,
) -> List[Any]:
    """
    Read the dbt manifest and import models as datasets with metrics.
    """
    base_url = URL(external_url_prefix) if external_url_prefix else None

    # add datasets
    datasets = []
    for model in models:
        filters = {
            "database": OneToMany(database["id"]),
            "schema": model["schema"],
            "table_name": model["name"],
        }
        existing = client.get_datasets(**filters)
        if len(existing) > 1:
            raise Exception("More than one dataset found")

        if existing:
            dataset = existing[0]
            _logger.info("Updating dataset %s", model["unique_id"])
        else:
            _logger.info("Creating dataset %s", model["unique_id"])
            try:
                dataset = client.create_dataset(
                    database=database["id"],
                    schema=model["schema"],
                    table_name=model["name"],
                )
            except Exception:  # pylint: disable=broad-except
                # Superset can't add tables from different BigQuery projects
                continue

        extra = {
            "unique_id": model["unique_id"],
            "depends_on": "ref('{name}')".format(**model),
            "certification": {
                "details": "This table is produced by dbt",
            },
        }

        dataset_metrics = []
        model_metrics = {
            metric["name"]: metric
            for metric in metrics
            if model["unique_id"] in metric["depends_on"]
        }
        for name, metric in model_metrics.items():
            dataset_metrics.append(
                {
                    "expression": get_metric_expression(name, model_metrics),
                    "metric_name": name,
                    "metric_type": metric["type"],
                    "verbose_name": name,
                    "description": metric["description"],
                    **metric["meta"],
                },
            )

        # update dataset clearing metrics...
        update = {
            "description": model["description"],
            "extra": json.dumps(extra),
            "is_managed_externally": disallow_edits,
            "metrics": [],
        }
        update.update(model.get("meta", {}).get("superset", {}))
        if base_url:
            fragment = "!/model/{unique_id}".format(**model)
            update["external_url"] = str(base_url.with_fragment(fragment))
        client.update_dataset(dataset["id"], **update)

        # ...then update metrics
        if dataset_metrics:
            update = {
                "metrics": dataset_metrics,
            }
            client.update_dataset(dataset["id"], **update)

        datasets.append(dataset)

    return datasets



"""
Tests for ``preset_cli.cli.superset.sync.dbt.databases``.
"""
# pylint: disable=invalid-name

from pathlib import Path

import pytest
import yaml
from pyfakefs.fake_filesystem import FakeFilesystem
from pytest_mock import MockerFixture

from preset_cli.cli.superset.sync.dbt.databases import sync_database
from preset_cli.exceptions import DatabaseNotFoundError


def test_sync_database_new(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test ``sync_database`` when we want to import a new DB.
    """
    fs.create_file(
        "/path/to/.dbt/profiles.yml",
        contents=yaml.dump({"my_project": {"outputs": {"dev": {}}}}),
    )
    mocker.patch(
        "preset_cli.cli.superset.sync.dbt.databases.build_sqlalchemy_params",
        return_value={"sqlalchemy_uri": "dummy://"},
    )
    client = mocker.MagicMock()
    client.get_databases.return_value = []

    sync_database(
        client=client,
        profiles_path=Path("/path/to/.dbt/profiles.yml"),
        project_name="my_project",
        target_name="dev",
        import_db=True,
        disallow_edits=False,
        external_url_prefix="",
    )

    client.create_database.assert_called_with(
        database_name="my_project_dev",
        sqlalchemy_uri="dummy://",
        is_managed_externally=False,
    )


def test_sync_database_new_custom_sqlalchemy_uri(
    mocker: MockerFixture,
    fs: FakeFilesystem,
) -> None:
    """
    Test ``sync_database`` when we want to import a new DB.
    """
    fs.create_file(
        "/path/to/.dbt/profiles.yml",
        contents=yaml.dump(
            {
                "my_project": {
                    "outputs": {
                        "dev": {
                            "meta": {
                                "superset": {
                                    "connection_params": {
                                        "sqlalchemy_uri": "sqlite://",
                                    },
                                    "database_name": "my_database",
                                },
                            },
                        },
                    },
                },
            },
        ),
    )
    mocker.patch(
        "preset_cli.cli.superset.sync.dbt.databases.build_sqlalchemy_params",
        return_value={"sqlalchemy_uri": "dummy://"},
    )
    client = mocker.MagicMock()
    client.get_databases.return_value = []

    sync_database(
        client=client,
        profiles_path=Path("/path/to/.dbt/profiles.yml"),
        project_name="my_project",
        target_name="dev",
        import_db=True,
        disallow_edits=False,
        external_url_prefix="",
    )

    client.create_database.assert_called_with(
        database_name="my_database",
        sqlalchemy_uri="sqlite://",
        is_managed_externally=False,
    )


def test_sync_database_env_var(
    mocker: MockerFixture,
    fs: FakeFilesystem,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """
    Test ``sync_database`` when the profiles file uses ``env_var``.
    """
    monkeypatch.setenv("dsn", "sqlite://")

    fs.create_file(
        "/path/to/.dbt/profiles.yml",
        contents=yaml.dump(
            {
                "my_project": {
                    "outputs": {
                        "dev": {
                            "meta": {
                                "superset": {
                                    "connection_params": {
                                        "sqlalchemy_uri": '{{ env_var("dsn") }}',
                                    },
                                    "database_name": "my_database",
                                },
                            },
                        },
                    },
                },
            },
        ),
    )
    mocker.patch(
        "preset_cli.cli.superset.sync.dbt.databases.build_sqlalchemy_params",
        return_value={"sqlalchemy_uri": "dummy://"},
    )
    client = mocker.MagicMock()
    client.get_databases.return_value = []

    sync_database(
        client=client,
        profiles_path=Path("/path/to/.dbt/profiles.yml"),
        project_name="my_project",
        target_name="dev",
        import_db=True,
        disallow_edits=False,
        external_url_prefix="",
    )

    client.create_database.assert_called_with(
        database_name="my_database",
        sqlalchemy_uri="sqlite://",
        is_managed_externally=False,
    )


def test_sync_database_no_project(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test ``sync_database`` when the project is invalid.
    """
    fs.create_file(
        "/path/to/.dbt/profiles.yml",
        contents=yaml.dump({"my_project": {"outputs": {"dev": {}}}}),
    )
    client = mocker.MagicMock()
    client.get_databases.return_value = []

    with pytest.raises(Exception) as excinfo:
        sync_database(
            client=client,
            profiles_path=Path("/path/to/.dbt/profiles.yml"),
            project_name="my_other_project",
            target_name="dev",
            import_db=True,
            disallow_edits=False,
            external_url_prefix="",
        )
    assert (
        str(excinfo.value)
        == "Project my_other_project not found in /path/to/.dbt/profiles.yml"
    )


def test_sync_database_no_target(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test ``sync_database`` when the target is invalid.
    """
    fs.create_file(
        "/path/to/.dbt/profiles.yml",
        contents=yaml.dump({"my_project": {"outputs": {"dev": {}}}}),
    )
    client = mocker.MagicMock()
    client.get_databases.return_value = []

    with pytest.raises(Exception) as excinfo:
        sync_database(
            client=client,
            profiles_path=Path("/path/to/.dbt/profiles.yml"),
            project_name="my_project",
            target_name="prod",
            import_db=True,
            disallow_edits=False,
            external_url_prefix="",
        )
    assert (
        str(excinfo.value)
        == "Target prod not found in the outputs of /path/to/.dbt/profiles.yml"
    )


def test_sync_database_multiple_databases(
    mocker: MockerFixture,
    fs: FakeFilesystem,
) -> None:
    """
    Test ``sync_database`` when multiple databases are found.

    This should not happen, since database names are unique.
    """
    fs.create_file(
        "/path/to/.dbt/profiles.yml",
        contents=yaml.dump({"my_project": {"outputs": {"dev": {}}}}),
    )
    mocker.patch(
        "preset_cli.cli.superset.sync.dbt.databases.build_sqlalchemy_params",
        return_value={"sqlalchemy_uri": "dummy://"},
    )
    client = mocker.MagicMock()
    client.get_databases.return_value = [
        {"id": 1, "database_name": "my_project_dev", "sqlalchemy_uri": "dummy://"},
        {"id": 2, "database_name": "my_project_dev", "sqlalchemy_uri": "dummy://"},
    ]

    with pytest.raises(Exception) as excinfo:
        sync_database(
            client=client,
            profiles_path=Path("/path/to/.dbt/profiles.yml"),
            project_name="my_project",
            target_name="dev",
            import_db=True,
            disallow_edits=False,
            external_url_prefix="",
        )
    assert str(excinfo.value) == "More than one database with the same name found"


def test_sync_database_external_url_prefix(
    mocker: MockerFixture,
    fs: FakeFilesystem,
) -> None:
    """
    Test ``sync_database`` with an external URL prefix.
    """
    fs.create_file(
        "/path/to/.dbt/profiles.yml",
        contents=yaml.dump({"my_project": {"outputs": {"dev": {}}}}),
    )
    mocker.patch(
        "preset_cli.cli.superset.sync.dbt.databases.build_sqlalchemy_params",
        return_value={"sqlalchemy_uri": "dummy://"},
    )
    client = mocker.MagicMock()
    client.get_databases.return_value = []

    sync_database(
        client=client,
        profiles_path=Path("/path/to/.dbt/profiles.yml"),
        project_name="my_project",
        target_name="dev",
        import_db=True,
        disallow_edits=True,
        external_url_prefix="https://dbt.example.org/",
    )

    client.create_database.assert_called_with(
        database_name="my_project_dev",
        sqlalchemy_uri="dummy://",
        external_url="https://dbt.example.org/#!/overview",
        is_managed_externally=True,
    )


def test_sync_database_existing(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test ``sync_database`` when we want to import an existing DB.
    """
    fs.create_file(
        "/path/to/.dbt/profiles.yml",
        contents=yaml.dump({"my_project": {"outputs": {"dev": {}}}}),
    )
    mocker.patch(
        "preset_cli.cli.superset.sync.dbt.databases.build_sqlalchemy_params",
        return_value={"sqlalchemy_uri": "dummy://"},
    )
    client = mocker.MagicMock()
    client.get_databases.return_value = [
        {"id": 1, "database_name": "my_project_dev", "sqlalchemy_uri": "dummy://"},
    ]

    sync_database(
        client=client,
        profiles_path=Path("/path/to/.dbt/profiles.yml"),
        project_name="my_project",
        target_name="dev",
        import_db=True,
        disallow_edits=False,
        external_url_prefix="",
    )

    client.update_database.assert_called_with(
        database_id=1,
        database_name="my_project_dev",
        is_managed_externally=False,
    )


def test_sync_database_new_no_import(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test ``sync_database`` when we want to import a new DB.
    """
    fs.create_file(
        "/path/to/.dbt/profiles.yml",
        contents=yaml.dump({"my_project": {"outputs": {"dev": {}}}}),
    )
    mocker.patch(
        "preset_cli.cli.superset.sync.dbt.databases.build_sqlalchemy_params",
        return_value={"sqlalchemy_uri": "dummy://"},
    )
    client = mocker.MagicMock()
    client.get_databases.return_value = []

    with pytest.raises(DatabaseNotFoundError):
        sync_database(
            client=client,
            profiles_path=Path("/path/to/.dbt/profiles.yml"),
            project_name="my_project",
            target_name="dev",
            import_db=False,
            disallow_edits=False,
            external_url_prefix="",
        )



"""
Commands to import RLS rules, ownership, and more.
"""

import click
import yaml
from yarl import URL

from preset_cli.api.clients.superset import SupersetClient


@click.command()
@click.argument(
    "path",
    type=click.Path(resolve_path=True),
    default="rls.yaml",
)
@click.pass_context
def import_rls(ctx: click.core.Context, path: str) -> None:
    """
    Import RLS rules from a YAML file.
    """
    auth = ctx.obj["AUTH"]
    url = URL(ctx.obj["INSTANCE"])
    client = SupersetClient(url, auth)

    with open(path, encoding="utf-8") as input_:
        config = yaml.load(input_, Loader=yaml.SafeLoader)
        for rls in config:
            client.import_rls(rls)


@click.command()
@click.argument(
    "path",
    type=click.Path(resolve_path=True),
    default="ownership.yaml",
)
@click.pass_context
def import_ownership(ctx: click.core.Context, path: str) -> None:
    """
    Import resource ownership from a YAML file.
    """
    auth = ctx.obj["AUTH"]
    url = URL(ctx.obj["INSTANCE"])
    client = SupersetClient(url, auth)

    with open(path, encoding="utf-8") as input_:
        config = yaml.load(input_, Loader=yaml.SafeLoader)
        for resource_name, ownership in config.items():
            client.import_ownership(resource_name, ownership)



"""
Helper functions.
"""

import ast
import json
import logging
import os
import re
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, TypedDict, Union

import yaml
from jinja2 import Environment
from sqlalchemy.engine.url import URL

from preset_cli.api.clients.dbt import ModelSchema

_logger = logging.getLogger(__name__)


def build_sqlalchemy_params(target: Dict[str, Any]) -> Dict[str, Any]:
    """
    Build the SQLAlchemy URI for a given target.
    """
    type_ = target.get("type")

    if type_ in {"postgres", "redshift"}:
        return build_postgres_sqlalchemy_params(target)
    if type_ == "bigquery":
        return build_bigquery_sqlalchemy_params(target)
    if type_ == "snowflake":
        return build_snowflake_sqlalchemy_params(target)

    raise NotImplementedError(
        f"Unable to build a SQLAlchemy URI for a target of type {type_}. Please file an "
        "issue at https://github.com/preset-io/backend-sdk/issues/new?labels=enhancement&"
        f"title=Backend+for+{type_}.",
    )


def build_postgres_sqlalchemy_params(target: Dict[str, Any]) -> Dict[str, Any]:
    """
    Build the SQLAlchemy URI for a Postgres target.
    """
    if "search_path" in target:
        _logger.warning("Specifying a search path is not supported in Apache Superset")

    username = target["user"]
    password = target.get("password") or target.get("pass")
    host = target["host"]
    port = target["port"]
    dbname = target["dbname"]

    query = {"sslmode": target["sslmode"]} if "sslmode" in target else None

    return {
        "sqlalchemy_uri": str(
            URL(
                drivername="postgresql+psycopg2",
                username=username,
                password=password,
                host=host,
                port=port,
                database=dbname,
                query=query,
            ),
        ),
    }


def build_bigquery_sqlalchemy_params(target: Dict[str, Any]) -> Dict[str, Any]:
    """
    Build the SQLAlchemy URI for a BigQuery target.

    Currently supports only configuration via ``keyfile``.
    """
    parameters: Dict[str, Any] = {}

    parameter_map = {
        "priority": "priority",
        "location": "location",
        "maximum_bytes_billed": "maximum_bytes_billed",
    }
    query = {
        kwarg: str(target[key]) for kwarg, key in parameter_map.items() if key in target
    }
    if "priority" in query:
        query["priority"] = query["priority"].upper()
    parameters["sqlalchemy_uri"] = str(
        URL(
            drivername="bigquery",
            host=target["project"],
            database="",
            query=query,
        ),
    )

    if "keyfile" not in target:
        raise Exception(
            "Only service account auth is supported, you MUST pass `keyfile`.",
        )

    with open(target["keyfile"], encoding="utf-8") as input_:
        credentials_info = json.load(input_)
        parameters["encrypted_extra"] = json.dumps(
            {"credentials_info": credentials_info},
        )

    return parameters


def build_snowflake_sqlalchemy_params(target: Dict[str, Any]) -> Dict[str, Any]:
    """
    Build the SQLAlchemy URI for a Snowflake target.
    """
    username = target["user"]
    password = target["password"] or None
    database = target["database"]
    host = target["account"]
    query = {"role": target["role"], "warehouse": target["warehouse"]}

    return {
        "sqlalchemy_uri": str(
            URL(
                drivername="snowflake",
                username=username,
                password=password,
                host=host,
                database=database,
                query=query,
            ),
        ),
    }


def env_var(var: str, default: Optional[str] = None) -> str:
    """
    Simplified version of dbt's ``env_var``.

    We need this to load the profile with secrets.
    """
    if var not in os.environ and not default:
        raise Exception(f"Env var required but not provided: '{var}'")
    return os.environ.get(var, default or "")


def as_number(value: str) -> Union[int, float]:
    """
    Simplified version of dbt's ``as_number``.
    """
    try:
        return int(value)
    except ValueError:
        return float(value)


class Target(TypedDict):
    """
    Information about the warehouse connection.
    """

    profile_name: str
    name: str
    schema: str
    type: str
    threads: int


def load_profiles(path: Path, project_name: str, target_name: str) -> Dict[str, Any]:
    """
    Load the file and apply Jinja2 templating.
    """
    with open(path, encoding="utf-8") as input_:
        profiles = yaml.load(input_, Loader=yaml.SafeLoader)

    if project_name not in profiles:
        raise Exception(f"Project {project_name} not found in {path}")
    project = profiles[project_name]
    outputs = project["outputs"]
    if target_name not in outputs:
        raise Exception(f"Target {target_name} not found in the outputs of {path}")
    target = outputs[target_name]

    env = Environment()
    env.filters["as_bool"] = bool
    env.filters["as_native"] = ast.literal_eval
    env.filters["as_number"] = as_number
    env.filters["as_text"] = str

    context = {
        "env_var": env_var,
        "project_name": project_name,
        "target": target,
    }

    def apply_templating(config: Any) -> Any:
        """
        Apply Jinja2 templating to dictionary values recursively.
        """
        if isinstance(config, dict):
            for key, value in config.items():
                config[key] = apply_templating(value)
        elif isinstance(config, list):
            config = [apply_templating(el) for el in config]
        elif isinstance(config, str):
            template = env.from_string(config)
            config = yaml.load(template.render(**context), Loader=yaml.SafeLoader)

        return config

    return apply_templating(profiles)


def filter_models(models: List[ModelSchema], condition: str) -> List[ModelSchema]:
    """
    Filter a list of dbt models given a select condition.

    Currently only a subset of the syntax is supported.

    See https://docs.getdbt.com/reference/node-selection/syntax.
    """
    # match by tag
    if condition.startswith("tag:"):
        tag = condition.split(":", 1)[1]
        return [model for model in models if tag in model["tags"]]

    # simple match by name
    model_names = {model["name"]: model for model in models}
    if condition in model_names:
        return [model_names[condition]]

    # plus and n-plus operators
    if "+" in condition:
        return filter_plus_operator(models, condition)

    # at operator -- from the docs it seems that it can only be used before the model name
    # (https://docs.getdbt.com/reference/node-selection/graph-operators#the-at-operator)
    if condition.startswith("@"):
        return filter_at_operator(models, condition)

    raise NotImplementedError(
        f"Unable to parse the selection {condition}. Please file an issue at "
        "https://github.com/preset-io/backend-sdk/issues/new?labels=enhancement&"
        f"title=dbt+select+{condition}.",
    )


def filter_plus_operator(
    models: List[ModelSchema],
    condition: str,
) -> List[ModelSchema]:
    """
    Filter a list of models using the plus or n-plus operators.
    """
    model_ids = {model["unique_id"]: model for model in models}
    model_names = {model["name"]: model for model in models}

    match = re.match(r"^(\d*\+)?(.*?)(\+\d*)?$", condition)
    # pylint: disable=invalid-name
    up, name, down = match.groups()  # type: ignore
    base_model = model_names[name]
    selected_models: Dict[str, ModelSchema] = {}

    if up:
        degrees = None if len(up) == 1 else int(up[:-1])
        queue = [(base_model, 0)]
        while queue:
            model, degree = queue.pop(0)
            id_ = model["unique_id"]
            if id_ not in selected_models:
                selected_models[id_] = model
                if degrees is None or degree < degrees:
                    queue.extend(
                        (model_ids[parent_id], degree + 1)
                        for parent_id in model["depends_on"]
                        if parent_id in model_ids
                    )

    if down:
        degrees = None if len(down) == 1 else int(down[1:])
        queue = [(base_model, 0)]
        while queue:
            model, degree = queue.pop(0)
            id_ = model["unique_id"]
            if id_ not in selected_models:
                selected_models[id_] = model
                if degrees is None or degree < degrees:
                    queue.extend(
                        (model_ids[child_id], degree + 1)
                        for child_id in model["children"]
                        if child_id in model_ids
                    )

    return list(selected_models.values())


def filter_at_operator(models: List[ModelSchema], condition: str) -> List[ModelSchema]:
    """
    filter a list of models using the at operator.
    """
    model_ids = {model["unique_id"]: model for model in models}
    model_names = {model["name"]: model for model in models}

    base_model = model_names[condition[1:]]
    selected_models: Dict[str, ModelSchema] = {}

    queue = [base_model]
    while queue:
        model = queue.pop(0)
        id_ = model["unique_id"]
        if id_ not in selected_models:
            selected_models[id_] = model

            # add children
            queue.extend(
                model_ids[child_id]
                for child_id in model["children"]
                if child_id in model_ids
            )

            # add parents of the children of the selected model
            if model != base_model:
                queue.extend(
                    model_ids[parent_id]
                    for parent_id in model["depends_on"]
                    if parent_id in model_ids
                )

    return list(selected_models.values())


def apply_select(
    models: List[ModelSchema],
    select: Tuple[str, ...],
    exclude: Tuple[str, ...],
) -> List[ModelSchema]:
    """
    Apply dbt node selection (https://docs.getdbt.com/reference/node-selection/syntax).
    """
    if not select:
        return models

    model_ids = {model["unique_id"]: model for model in models}

    selected: Dict[str, ModelSchema] = {}
    for selection in select:
        ids = set.intersection(
            *[
                {model["unique_id"] for model in filter_models(models, condition)}
                for condition in selection.split(",")
            ]
        )
        selected.update({id_: model_ids[id_] for id_ in ids})

    for selection in exclude:
        for id_ in set.intersection(
            *[
                {model["unique_id"] for model in filter_models(models, condition)}
                for condition in selection.split(",")
            ]
        ):
            del selected[id_]

    return list(selected.values())



"""
Test JWT auth.
"""

import pytest
from pytest_mock import MockerFixture

from preset_cli.auth.jwt import JWTAuth


def test_jwt_auth() -> None:
    """
    Test the ``JWTAuth`` authentication mechanism.
    """
    auth = JWTAuth("my-token")
    assert auth.get_headers() == {"Authorization": "Bearer my-token"}


def test_jwt_auth_from_stored_credentials(mocker: MockerFixture) -> None:
    """
    Test instantiating the object from stored credentials
    """
    mocker.patch("preset_cli.auth.jwt.open")

    get_credentials_path = mocker.patch("preset_cli.auth.jwt.get_credentials_path")
    get_credentials_path().exists.return_value = True
    get_credentials_path().__str__.return_value = "/path/to/credentials.yaml"

    yaml = mocker.patch("preset_cli.auth.jwt.yaml")
    yaml.load.return_value = {
        "api_token": "TOKEN",
        "api_secret": "SECRET",
        "baseurl": "https://manage.app.preset.io/",
    }

    get_access_token = mocker.patch("preset_cli.auth.jwt.get_access_token")
    get_access_token.return_value = "JWT_TOKEN"

    auth = JWTAuth.from_stored_credentials()
    assert auth.token == "JWT_TOKEN"
    get_access_token.assert_called_with(
        baseurl="https://manage.app.preset.io/",
        api_token="TOKEN",
        api_secret="SECRET",
    )

    # can also pass a URL
    auth = JWTAuth.from_stored_credentials()
    assert auth.token == "JWT_TOKEN"

    # test for error
    get_credentials_path().exists.return_value = False
    with pytest.raises(Exception) as excinfo:
        JWTAuth.from_stored_credentials()
    assert (
        str(excinfo.value)
        == "Could not load credentials from /path/to/credentials.yaml"
    )



"""
Tests for ``preset_cli.api.clients.preset``.
"""

from requests_mock.mocker import Mocker

from preset_cli.api.clients.preset import PresetClient
from preset_cli.auth.main import Auth


def test_preset_client_get_teams(requests_mock: Mocker) -> None:
    """
    Test the ``get_teams`` method.
    """
    auth = Auth()
    requests_mock.get("https://ws.preset.io/api/v1/teams/", json={"payload": [1, 2, 3]})

    client = PresetClient("https://ws.preset.io/", auth)
    teams = client.get_teams()
    assert teams == [1, 2, 3]


def test_preset_client_get_workspaces(requests_mock: Mocker) -> None:
    """
    Test the ``get_workspaces`` method.
    """
    auth = Auth()
    requests_mock.get(
        "https://ws.preset.io/api/v1/teams/botafogo/workspaces/",
        json={"payload": [1, 2, 3]},
    )

    client = PresetClient("https://ws.preset.io/", auth)
    teams = client.get_workspaces("botafogo")
    assert teams == [1, 2, 3]
